package combinatorParsing

/** A class of generic parsers
 *  <p>
 *    This class defines:
 *  </p>
 *  <ul>
 *    <li>a type Parser.</li>
 *    <li>generic parser combinators: sequence, alternative, result production,
 *      option, repetition, etc.</li>
 *    <li>primitive parser generators for the token classes, success and failure.</li>
 *  </ul>
 *  <p>
 */
trait GenericParsers extends Tokens {

  /** The parser input is a list of tokens */
  type Input = List[Token]

  /** A base class for parser results. 
   *  Results are either successes or failures (see below).
   */
  class ParseResult

  /** The success case of ParseResult.
   *
   *  @param out    The parser's output (which is a list of terminals or non-terminals)
   *  @param in     The parser's remaining input
   */
  case class Success(out: List[Token], in: Input) extends ParseResult {
    /** The toString method of a Success yields each result token, separated by newlines */
    override def toString = out.mkString("", "\n", "")
  }

  /** The failure case of ParseResult.
   *
   *  @param msg    An error message string describing the failure
   *  @param in     The parser's unconsumed input at the point where the failure occured
   */
  case class Failure(msg: String, in: Input) extends ParseResult {
    /** The toString method of a Failure yields an error message */
    override def toString = "["+in.head.pos+"] error: "+msg+"\n\n"+in.head.pos.longString    
  }

  /** The root class of parsers. 
   *  Parsers are functions from the Input type to ParseResult 
   */
  abstract class Parser extends (Input => ParseResult) {

    /** An unspecified method the defines the parser function.
     */
    def apply(in: Input): ParseResult

    /** A parser combinator for sequential composition 
     *
     *  `p & q' succeeds if `p' succeeds and `q' succeeds on the input
     *          left over by `p'.
     */
    def & (q: => Parser) = new Parser {
      def apply(in: Input) = Parser.this(in) match {
        case Success(x, in1) => 
          q(in1) match {
            case Success(y, in2) => Success(x ::: y, in2)
            case failure => failure
          }
        case failure => failure
      }
    }

    /** <p>
     *    A parser combinator for alternative composition 
     *  </p>
     *  <p>
     *    <code>p | q</code> succeeds if <code>p</code> succeeds or
     *    <code>q</code> succeeds.
     *  </p>
     */
    def | (q: => Parser) = new Parser {
      def apply(in: Input) = Parser.this(in) match {
        case s1 @ Success(_, _) => s1
        case f1 @ Failure(_, in1) => q(in) match {
          case s2 @ Success(_, _) => s2
          case f2 @ Failure(_, in2) => if (in2.head.pos < in1.head.pos) f1 else f2
        }
      }
    }
  }

  /** A parser generator delimiting whole productions 
   *
   *  production(p) succeeds if p succeeds. In this case, production(p) will
   *  generate an output value from the list of tokens produced by `p'.
   */
  def production(p: Parser) = new Parser {
    def apply(in: Input) = p(in)
  }

  /** A parser generator delimiting whole phrases (i.e. programs).
   *  phrase(p) succeeds if `p' succeeds and no input is left over after `p'.
   */
  def phrase(p: Parser) = new Parser {
    def apply(in: Input) = p(in) match {
      case s @ Success(out, List(EOF())) => s
      case s @ Success(out, in1) => Failure("end of input expected", in1)
      case f @ Failure(_, in @ (ErrorToken(msg) :: _)) => Failure(msg, in)
      case f => f
    }
  }

  /** <p>
   *    A parser matching tokens which satisfy a given predicate.
   *  </p>
   *  <p>
   *    <code>token(kind, p)</code> succeeds if the input starts with a token
   *    <code>t</code> for which <code>p(t)</code> is true.
   *  </p>
   *
   *  @param  kind   The token's kind, used for error messages
   *  @param  p      A predicate which determins which tokens match.
   *  @return        A parser matching the specified tokens.
   *
   */
  def token(kind: String, p: Token => boolean) = new Parser {
    def apply(in: Input) =
      if (p(in.head)) Success(List(in.head), in.tail)
      else Failure(kind+" expected", in)
  }

  /** A parser which matches a single keyword token.
   *  @param chars    The character string making up the matched keyword. */
  def kw(chars: String) = token("'"+chars+"'", {
    case KW(chars1) => chars == chars1
    case _ => false
  })

  /** A parser which matches a numeric literal */
  def numericLit = token("number", .isInstanceOf[NumericLit])

  /** A parser which matches a string literal */
  def stringLit = token("string literal", .isInstanceOf[StringLit])

  /** A parser which matches an identifier */
  def ident = token("identifier", .isInstanceOf[Identifier])

  /** A parser which always fails.
   *
   *  @param msg   The error message describing the failure.
   *  @return      A parser which always fails.
   */
  def failure(msg: String) = new Parser {
    def apply(in: Input) = Failure(msg, in)
  }

  /** A parser which always succeeds.
   *
   *  @param out   The output value of the parser.
   *  @return      A parser which always succeeds.
   */
  def success(out: List[Token]) = new Parser {
    def apply(in: Input) = Success(out, in)
  }

  /** The parser which matches the empty input string
   *  and which returns the empty output list.
   */
  def empty = success(List())

  /** <p>
   *    A parser generator for repetitions.
   *  </p>
   *  <p>
   *    <code>rep(p)</code> runs <code>p</code> zero or more times on input
   *    until no more match.
   *  </p>
   */
  def rep(p: Parser): Parser = p & rep(p) | empty

  /** <p>
   *    A parser generator for non-empty repetitions.
   *  </p>
   *  <p>
   *    <code>rep1(p)</code> runs <code>p</code> one or more times on input
   *    until no more match.
   *  </p>
   */
  def rep1(p: Parser): Parser = p & rep(p)

  /** A parser generator for optional sub-phrases.
   *
   *  opt(p)   runs `p' zero or once on input
   */
  def opt(p: Parser): Parser = p | empty
}
